module BackgroundJobs
  module Sidekiq
    class JobAttributesAdapter
      def initialize(attributes)
        @attributes = Attributes.new(attributes)
      end

      def encode
        @attributes.encode
      end

      def decode
        @attributes.decode
      end
    end

    class Attributes < Array
      def initialize(array)
        super array.map {|attr| create_attribute(attr) }
      end

      def encode
        map(&:encode)
      end

      def decode
        map(&:decode)
      end

      private

      def create_attribute(attr)
        AttributeFactory.build(attr)
      end
    end

    class AttributeFactory
      def self.build(value)
        return HashAttribute.new(value) if value.kind_of?(Hash)
        return ArrayAttribute.new(value) if value.kind_of?(Array)
        return Attribute.new(value) unless value.kind_of?(String)
        StringAttribute.new(value)
      end
    end

    class Attribute
      def initialize(value)
        @value = value
      end

      def decode
        @value
      end

      def encode
        case @value.class.name
          when 'Date'
            "date:#{@value.to_s}"

          when 'Hash'
            HashAttribute.new(@value).encode

          when 'Array'
            ArrayAttribute.new(@value).encode

          when 'Range'
            "range:[#{Attribute.new(@value.first).encode},#{Attribute.new(@value.last).encode}]"

          else
            @value
        end
      end
    end

    class HashAttribute < Hash
      def initialize(value)
        @value = value
      end

      def encode
        @value.keys.each do |key|
          value = @value.delete(key)
          @value[key.to_s] = AttributeFactory.build(value).encode
        end
        
        @value
      end

      def decode
        @value.keys.each do |key|
          @value[key.to_sym] = AttributeFactory.build(@value.delete(key)).decode
        end

        @value
      end
    end

    class ArrayAttribute < Array
      def decode
        map {|i| AttributeFactory.build(i).decode }
      end

      def encode
        map {|i| AttributeFactory.build(i).encode }
      end
    end

    class StringAttribute < String
      def decode
        if number?
          return to_i
        end

        if date?
          return Date.parse(gsub('date:', ''))
        end

        if range?
          matches = match(/range:\[(.*),(.*)\]/)
          return AttributeFactory.build(matches[1]).decode..AttributeFactory.build(matches[2]).decode
        end

        self
      end

      private

      def number?
        match(/^\d*$/)
      end      

      def date?
        match(/^date:.*$/)
      end

      def range?
        match(/^range:\[(.*),(.*)\]$/)
      end
    end
  end
end